/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import java.io.Serializable;
import java.math.BigInteger;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
/**
 * Grid2DSquareCellChunkAbstract provides inner classes for ChunkCellID and
 * general geometry methods for extended classes. It also controls what methods 
 * extended classes must implement acting like an interface.
 *
 * The basic geometries are ordered in set numbers of rows and columns and are
 * arranged sequentially as their base two-dimensional orthogonal coordinate
 * axes. The sequential arrangement goes along the x-axis row by row from the
 * y-axis, then up the y-axis taking each row in turn.
 *
 * TODO:
 * Add a int TYPE attribute field which can be used to switch between different 
 * types of instances. 
 */
public abstract class Grid2DSquareCellChunkAbstract implements Serializable {
    
    private static final long serialVersionUID = 1L; 
    
    /**
     * A reference to the Grid2DSquareCellDoubleAbstract instance.
     */
    protected transient Grid2DSquareCellAbstract grid2DSquareCell;
    
    /**
     * For storing the ChunkID of this.
     * TODO:
     * Is this transient for caching?
     */
    protected transient ChunkID chunkID;
    
    /**
     * Indicator for whether the swapped version of this chunk is upToDate. This adds 
     * a small amount of weight, so for 64CellMap implementations it may be undesirable.
     */
    protected transient boolean isSwapUpToDate;
    
    /**
     * Returns this.grid2DSquareCell.
     */
    public Grid2DSquareCellAbstract getGrid2DSquareCell() {
        return this.grid2DSquareCell;
    }
    
    /**
     * Initialises grid2DSquareCell.
     * @param grid2DSquareCell
     */
    protected void initGrid2DSquareCell( 
            Grid2DSquareCellAbstract grid2DSquareCell ) {
        this.grid2DSquareCell = grid2DSquareCell;
    }
    
    /**
     * Initialises chunkID.
     * @param chunkID
     */
    protected void initChunkID( 
            ChunkID chunkID ) {
        this.chunkID = chunkID;
    }
    
    /**
     * Returns a copy of this.chunkID.
     * TODO:
     * Check that returning a clone is fine.
     */
    public ChunkID getChunkID() {
        return new ChunkID( this.chunkID );
        //return this.chunkID;
    }
    
    /**
     * Returns this.isSwapUpToDate
     */
    protected boolean getIsSwapUpToDate() {
        return this.isSwapUpToDate;
    }

    /**
     * Returns this.isSwapUpToDate
     */
    protected void setIsSwapUpToDate( 
            boolean isSwapUpToDate ) {
        this.isSwapUpToDate = isSwapUpToDate;
    }
    
    /**
     * For returning a descritption of this.
     */
    public abstract String toString();

    /**
     * Returns the name of this.
     */
    public String getName() {
        return this.getClass().getName();
    }
    
    /**
     * Returns an iterator over the cell values. These are not guaranteed to be in any particular order.
     */
    public abstract IteratorAbstract iterator();
    
    /**
     * Returns true if the cell given by chunk cell row index 
     * chunkCellRowIndex, chunk cell col index chunkCellColIndex is in this.
     * @param chunkCellRowIndex 
     * @param chunkCellColIndex
     */
    public boolean inChunk( 
            int chunkCellRowIndex, 
            int chunkCellColIndex ) {
        int chunkNrows = this.grid2DSquareCell.getChunkNrows( this.chunkID );
        int chunkNcols = this.grid2DSquareCell.getChunkNcols( this.chunkID );
        return ( 
                chunkCellRowIndex > -1 && 
                chunkCellRowIndex < chunkNrows && 
                chunkCellColIndex > -1 && 
                chunkCellColIndex < chunkNcols );
    }
    
//    /**
//     * Initialises the value of this with those in the randomAccessFile. The
//     * initialisation is done in row order and if the end of the file is reached
//     * then the remaining values are assumed to be noDataValues.
//     * NB1. This method is intended mainly for use in caching
//     */
//    protected abstract void initCells( Grid2DSquareCellChunkRAF grid2DSquareCellChunkRAF );
    
    /**
     * For clearing the data associated with this.
     */
    public abstract void clearData();
    
    /**
     * For initialising the data associated with this.
     */
    protected abstract void initData();
    
    /**
     * A simple ChunkCellID class for distinguishing cells in this chunk.
     */
    public class ChunkCellID
            implements Serializable, Comparable {

        /**
         * For storing the row in which the cell is positioned
         */
        protected int chunkCellRowIndex;

        /**
         * For storing the column in which the cell is positioned
         */
        protected int chunkCellColIndex;

        /**
         * For storing a hashCode for the cell.
         * NB. When the number of cells in a grid is more than the maximum int 
         * value care should be taken to distributed hashcodes in a relatively 
         * even way. Operations will slow down greatly if a large number of 
         * cells have a ChunkCellID with the same hashCode.
         */
        protected int hashCode;

        /**
         * Constructs a new ChunkCellID
         */
        protected ChunkCellID() {}
        
        /**
         * Constructs a new ChunkCellID
         * @param chunkCellRowIndex the row of the cell
         * @param chunkCellColIndex the column of the cell
         */
        protected ChunkCellID( 
                int chunkCellRowIndex, 
                int chunkCellColIndex ) {
            this.chunkCellRowIndex = chunkCellRowIndex;
            this.chunkCellColIndex = chunkCellColIndex;
            //this.hashCode = ( int ) ( ( getChunkNcols() * chunkCellRowIndex ) + chunkCellColIndex );
            //this.hashCode = ( int ) ( ( chunkNcols * chunkCellRowIndex ) + chunkCellColIndex );
            //this.hashCode = ( int ) ( ( getGrid2DSquareCell().getChunkNcols() * chunkCellRowIndex ) + chunkCellColIndex );
            this.hashCode = ( int ) ( ( getGrid2DSquareCell().chunkNcols * chunkCellRowIndex ) + chunkCellColIndex );
        }
        
        /**
         * Constructs a new ChunkCellID
         * @param chunkCellRowIndex the row of the cell
         * @param chunkCellColIndex the column of the cell
         * @param chunkNcols the number of columns in the chunk
         */
        protected ChunkCellID( 
                int chunkNcols, 
                int chunkCellRowIndex, 
                int chunkCellColIndex ) {
            this.chunkCellRowIndex = chunkCellRowIndex;
            this.chunkCellColIndex = chunkCellColIndex;
            this.hashCode = 
                    ( int ) ( ( chunkNcols * chunkCellRowIndex ) + chunkCellColIndex );
        }

        /**
         * Returns hashCode of this CellID
         */
        public int hashCode() {
            return this.hashCode;
        }
        
        /**
         * Returns chunkCellRowIndex of this CellID
         */
        public long getChunkCellRowIndex() {
            return this.chunkCellRowIndex;
        }
        
        /**
         * Returns chunkCellColIndex of this CellID
         */
        public long getChunkCellColIndex() {
            return this.chunkCellColIndex;
        }
        
        /**
         * Returns a description of this CellID
         */
        public String toString() {
            return  "ChunkCellID( hashCode( " + hashCode() + "), " +
                    "chunkCellRowIndex( " + getChunkCellRowIndex() + " ), " +
                    "chunkCellColIndex( " + getChunkCellColIndex() + " ) )";
        }
        
        /**
         * Overrides equals in Object
         */
        public boolean equals( Object object ) {
            if ( this == object ) {
                return true;
            }
            if ( ( object == null ) || ( object.getClass() != this.getClass() ) ) {
                return false;
            }
            ChunkCellID anotherChunkCellID = ( ChunkCellID ) object;
            return  ( this.chunkCellColIndex == anotherChunkCellID.chunkCellColIndex ) && 
                    ( this.chunkCellRowIndex == anotherChunkCellID.chunkCellRowIndex );
        }
        
        /**
         * Method required by Comparable
         */
        public int compareTo( Object object ) {
            //try {
            //Cast failure should result in NullPointerException or ClassCastException
                ChunkCellID chunkCellID = ( ChunkCellID ) object;
                if ( chunkCellID.chunkCellRowIndex > this.chunkCellRowIndex ) {
                    return 1;
                }
                if ( chunkCellID.chunkCellRowIndex < this.chunkCellRowIndex ) {
                    return -1;
                }
                if ( chunkCellID.chunkCellColIndex > this.chunkCellColIndex ) {
                    return 1;
                }
                if ( chunkCellID.chunkCellColIndex < this.chunkCellColIndex ) {
                    return -1;
                }
                return 0;
           //} catch ( Exception e0 ) {
           //    return -1;
           //}
        }
    }
}