/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import gnu.trove.TDoubleHashSet;
import gnu.trove.TDoubleLongHashMap;
import gnu.trove.TDoubleLongIterator;
import gnu.trove.TDoubleLongProcedure;
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Arrays;
import java.util.ConcurrentModificationException;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
import uk.ac.leeds.ccg.andyt.grids.core.GridStatisticsInterface;
import uk.ac.leeds.ccg.andyt.grids.utilities.UnsignedLong;
import uk.ac.leeds.ccg.andyt.grids.utilities.UnsignedLongPowersOf2;
/**
 * Grid2DSquareCellDoubleChunkAbstract extension that stores cell values in a 
 * TDoubleLongHashMap. The maximum size of this 
 * Grid2DSquareCellDoubleChunkAbstract is 64 cells.
 * TODO:
 * Adjudicate when this is useful. What type of data?
 */
public class Grid2DSquareCellDoubleChunk64CellMap 
        extends Grid2DSquareCellDoubleChunkAbstract 
        implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    /**
     * For storing values mapped to a binary encoded long. The long is a key
     * which indicates if the value is that at a given location. Both keys and
     * values are unique.
     */
    private TDoubleLongHashMap data;
    
    /**
     * Default constructor.
     */
    public Grid2DSquareCellDoubleChunk64CellMap() {
        this.chunkID = new ChunkID();
        this.data = new TDoubleLongHashMap();
        this.grid2DSquareCell = new Grid2DSquareCellDouble();
        this.isSwapUpToDate = false;
    }
    
    /**
     * TODO:
     * documentation
     * deal with case:
     *   ( grid2DSquareCellDouble.getChunkNrows() * grid2DSquareCellDouble.getChunkNcols() > 64 )
     */
    protected Grid2DSquareCellDoubleChunk64CellMap( 
            Grid2DSquareCellDouble grid2DSquareCellDouble,
            ChunkID chunkID ) {
        //initChunkID( chunkID );
        this.chunkID = chunkID;
        initGrid2DSquareCell( grid2DSquareCellDouble );
        Long nCellsInChunk = 
                ( long ) grid2DSquareCellDouble.chunkNrows *
                ( long ) grid2DSquareCellDouble.chunkNcols;
        if ( nCellsInChunk <= 64 ) {
            initData();
        } else {
            System.err.println(
                    "grid2DSquareCellDoubleChunk.chunkNrows * " +
                    "grid2DSquareCellDoubleChunk.chunkNcols > 64 " +
                    "in \n" + 
                    this.getClass().getName() +
                    ".Grid2DSquareCellDoubleChunk64CellMap( \n" +
                    "    Grid2DSquareCellDouble ( " + 
                    grid2DSquareCellDouble.getBasicDescription() + " ), \n" +
                    "    ChunkID ( " + chunkID + " ) );" );
            // Sample? What to do? Hmmmmmmm!
        }
        this.isSwapUpToDate = false;
    }
    
    /**
     * TODO: docs
     */
    protected Grid2DSquareCellDoubleChunk64CellMap( 
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk, 
            ChunkID chunkID ) {
        this.chunkID = chunkID;
        Grid2DSquareCellDouble grid2DSquareCellDouble = 
                grid2DSquareCellDoubleChunk.getGrid2DSquareCellDouble();
        initGrid2DSquareCell( grid2DSquareCellDouble );
        int chunkNrows = grid2DSquareCellDouble.chunkNrows;
        int chunkNcols = grid2DSquareCellDouble.chunkNcols;
        if ( chunkNrows * chunkNcols <= 64 ) {
            double noDataValue = grid2DSquareCellDouble.noDataValue;
            initData();
            double value;
            for ( int row = 0; row < chunkNrows; row ++ ) {
                for ( int col = 0; col < chunkNcols; col ++ ) {
                    value = grid2DSquareCellDoubleChunk.getCell( 
                            row, 
                            col, 
                            noDataValue );
                    if ( value != noDataValue ) {
                        initCell( 
                                row, 
                                col, 
                                value );
                    }
                }
            }
        } else {
            System.err.println(
                    "grid2DSquareCellDoubleChunk.chunkNrows * " +
                    "grid2DSquareCellDoubleChunk.chunkNcols > 64 " +
                    "in \n" + 
                    this.getClass().getName() +
                    ".Grid2DSquareCellDoubleChunk64CellMap( \n" +
                    "    Grid2DSquareCellDouble ( " + 
                    grid2DSquareCellDouble.getBasicDescription() + " ), \n" +
                    "    ChunkID ( " + chunkID + " ) );" );
            // Sample? What to do? Hmmmmmmm!
        }
        this.isSwapUpToDate = false;
    }
    
    /**
     * Returns a description of this.
     */
    public String toString() {
        return  "Grid2DSquareCellDoubleChunk64CellMap( \n" +
                "ChunkID( " + this.chunkID.toString() + " ) )";
    }
    
    /**
     * Initialises the data associated with this.
     */
    public void initData() {
        this.data = new TDoubleLongHashMap();
    }
    
    //    /**
    //     * TODO: docs
    //     */
    //    public TDoubleLongHashMap getData(){
    //        return this.data;
    //    }
    
    /**
     * Clears the data associated with this.
     */
    public void clearData() {
        this.data = null;
        System.gc();
    }
    
    /**
     * Returns this.data
     */
    public TDoubleLongHashMap getData() {
        return this.data;
    }
    
    /**
     * TODO:
     * docs
     */
    public @Override double[] toArrayIncludingNoDataValues() {
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
        double noDataValue = grid2DSquareCellDouble.noDataValue;
        int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
        double[] array = new double[ nrows * ncols ];
        Arrays.fill( array, noDataValue );
        TDoubleLongIterator iterator = this.data.iterator();
        UnsignedLong valueMap = new UnsignedLong();
        int ite;
        boolean[] valueMapped;
        int position;
        for ( ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            valueMap.setLong( iterator.value() );
            valueMapped = valueMap.toBooleanArray();
            for ( position = 0; position < valueMapped.length; position ++ ) {
                if ( valueMapped[ position ] ) {
                    array[ position ] = iterator.key();
                }
            }
        }
        return array;
    }
    
    /**
     * TODO:
     * docs
     */
    public @Override double[] toArrayNotIncludingNoDataValues() {
        double[] array = new double[ getNonNoDataValueCountInt() ];
        TDoubleLongIterator iterator = this.data.iterator();
        UnsignedLong valueMap = new UnsignedLong();
        int ite;
        boolean[] valueMapped;
        int position;
        int index = 0;
        for ( ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            valueMap.setLong( iterator.value() );
            valueMapped = valueMap.toBooleanArray();
            for ( position = 0; position < valueMapped.length; position ++ ) {
                if ( valueMapped[ position ] ) {
                    array[ index ] = iterator.key();
                    index ++;
                }
            }
        }
        return array;
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex;
     * chunk cell row chunkCellColIndex.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param noDataValue the noDataValue of this.grid2DSquareCellDouble
     */
    public double getCell(
            int chunkCellRowIndex, 
            int chunkCellColIndex, 
            double noDataValue ) {
        TDoubleLongIterator iterator = this.data.iterator();
        UnsignedLong valueMap = new UnsignedLong();
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
        int chunkNcols = grid2DSquareCellDouble.chunkNcols;
        int ite;
        int position;
        for ( ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            valueMap.setLong( iterator.value() );
            position = ( chunkCellRowIndex * chunkNcols ) + chunkCellColIndex;
            if ( valueMap.isAtPosition( position ) ) {
                return iterator.key();
            }
        }
        return noDataValue;
    }
    
    /**
     * Initialises the value at position given by: chunk cell row chunkCellRowIndex;
     * chunk cell column chunkCellColIndex. Utility method for constructor.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToInitialise the value with which the cell is initialised
     */
    protected void initCell(
            int chunkCellRowIndex, 
            int chunkCellColIndex, 
            double valueToInitialise ) {
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
        int chunkNcols = grid2DSquareCellDouble.chunkNcols;
        long long0;
        if ( this.data.containsKey( valueToInitialise ) ) {
            long0 = data.get( valueToInitialise ) + 
                    powerOf2( ( ( chunkCellRowIndex * chunkNcols ) + chunkCellColIndex ) );
            this.data.put(
                    valueToInitialise,
                    long0 );
                     } else {
            long0 = powerOf2( ( ( chunkCellRowIndex * chunkNcols ) + chunkCellColIndex ) );
            this.data.put( 
                    valueToInitialise, 
                    long0 );
        }
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex;
     * chunk cell column chunkCellColIndex and sets it to valueToSet
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToSet the value the cell is to be set to
     * @param noDataValue the noDataValue of this.grid2DSquareCellDouble
     */
    public double setCell( 
            int chunkCellRowIndex, 
            int chunkCellColIndex, 
            double valueToSet, 
            double noDataValue ) {
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
        double result = noDataValue;
        TDoubleLongIterator iterator = this.data.iterator();
        UnsignedLong valueMap = new UnsignedLong();
        double value;
        boolean gotValue = false;
        boolean setValue = false;
        if ( valueToSet == result ) {
            setValue = true;
        } else {
            setValue = false;
        }
        int position = ( chunkCellRowIndex * grid2DSquareCellDouble.chunkNcols ) + chunkCellColIndex;
        long powerOf2 = powerOf2( position );
        int ite;
        for ( ite = 0; ite < data.size(); ite ++ ) {
            if ( ! ( gotValue && setValue ) ) {
                try{
                    iterator.advance();
                } catch ( ConcurrentModificationException why ) {
                 ite = ite; // debug?
                }
                valueMap.setLong( iterator.value() );
                value = iterator.key();
                //if ( ! gotValue ) { // Probably slows things down!?
                if ( valueMap.isAtPosition( position ) ) {
                    // May be better to keep the value in data and tidy if
                    // necessary. This would save evaluating the if, but would
                    // mean that the chunk is likely to get heavier. However it
                    // is difficult to know what is best, it probably depends on
                    // the data and what is being done with it!
                    if ( iterator.value() == powerOf2 ) {
                        if ( value == valueToSet ) {
                            return value;
                        } else {
                            //try {
                            this.data.remove( value ); // Removal causes ConcurrentModificationException in iterator.advance()!
                            //} catch ( ConcurrentModificationException e ) {}
                        }
                    } else {
                        iterator.setValue( iterator.value() - powerOf2 );
                    }
                    result = value;
                    gotValue = true;
                }
                //}
                //if ( ! setValue ) { // Probably slows things down!?
                if ( value == valueToSet ) {
                    iterator.setValue( iterator.value() + powerOf2 );
                    setValue = true;
                }
                //}
            } else {
                setIsSwapUpToDate( false );
                return result;
            }
        }
        if ( ! setValue ) {
            this.data.put( 
                    valueToSet,
                    powerOf2 );
        }
        setIsSwapUpToDate( false );
        return result;
    }
    
    /**
     * Returns 2 raised to the power of value as a long.
     * TODO: move this to the chunk master and have as a long[]. The look up
     * should be faster than the creation on the fly.
     */
    public long powerOf2( 
            int value ){
        Grid2DSquareCellDouble grid2DSquareCellDouble = this.getGrid2DSquareCellDouble();
        try {
            //return this.unsignedLongPowersOf2.powersOf2[ 63 - value ];
            return grid2DSquareCellDouble.unsignedLongPowersOf2.powersOf2[ value ];
        } catch ( java.lang.NullPointerException e ) {
            grid2DSquareCellDouble.unsignedLongPowersOf2 = new UnsignedLongPowersOf2( false );
            //return this.unsignedLongPowersOf2.powersOf2[ 63 - value ];
            return grid2DSquareCellDouble.unsignedLongPowersOf2.powersOf2[ value ];
        }
    }
    
    /**
     * Returns the number of cells with noDataValues as a BigInteger
     */
    public @Override BigInteger getNonNoDataValueCountBigInteger() {
        TDoubleLongIterator iterator = this.data.iterator();
        long cellIDMap = 0L;
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            cellIDMap += iterator.value();
        }
        return new BigInteger( "" + new UnsignedLong( cellIDMap ).getCount() );
    }
    
    /**
     * Returns the number of cells with noDataValues as an int
     */
    public @Override int getNonNoDataValueCountInt() {
        TDoubleLongIterator iterator = this.data.iterator();
        long cellIDMap = 0L;
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            cellIDMap += iterator.value();
        }
        return new UnsignedLong( cellIDMap ).getCount();
    }
    
    /**
     * For returning the sum of all non noDataValues as a BigDecimal
     */
    public @Override BigDecimal getSumBigDecimal() {
        BigDecimal sum = new BigDecimal( 0.0d );
        BigDecimal thisCount;
        TDoubleLongIterator iterator = this.data.iterator();
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = BigDecimal.valueOf( ( long ) value.getCount() );
            sum = sum.add( new BigDecimal( iterator.key() ).multiply( thisCount ) );
        }
        return sum;
    }
    
    /**
     * For returning the minimum of all non noDataValues as a double
     */
    public @Override double getMinDouble() {
        double min = Double.POSITIVE_INFINITY;
        TDoubleLongIterator iterator = this.data.iterator();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            min = Math.min( min, iterator.key() );
        }
        return min;
    }
    
    /**
     * For returning the maximum of all non noDataValues as a double
     */
    public @Override double getMaxDouble() {
        double max = Double.NEGATIVE_INFINITY;
        TDoubleLongIterator iterator = this.data.iterator();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            max = Math.max( max, iterator.key() );
        }
        return max;
    }
    
    /**
     * For returning the Arithmetic Mean of all non noDataValues as a BigDecimal.
     * Using BigDecimal this should be as precise as possible with doubles.
     * @param numberOfDecimalPlaces The number of decimal places for precision.
     */
    public @Override BigDecimal getArithmeticMeanBigDecimal( 
            int numberOfDecimalPlaces ) {
        BigDecimal mean = new BigDecimal( "0" );
        BigDecimal count = new BigDecimal( "0" );
        BigDecimal thisCount;
        TDoubleLongIterator iterator = this.data.iterator();
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = BigDecimal.valueOf( ( long ) value.getCount() );
            count = count.add( thisCount );
            mean = mean.add( new BigDecimal( iterator.key() ).multiply( thisCount ) );
        }
        // Might cause a divide by zero problem!
        try {
            return mean.divide( count, numberOfDecimalPlaces, BigDecimal.ROUND_HALF_EVEN );
        } catch ( ArithmeticException e ) {
            System.out.println(
                    "Divide by zero in \n" + 
                    this.getClass().getName() + ".getArithmeticMeanBigDecimal(). " +
                    "Returning null!" );
            return null;
        }
    }
    
    /**
     * For returning the Arithmetic Mean of all non noDataValues as a double.
     * Using BigDecimal this should be as precise as possible with doubles.
     * @param numberOfDecimalPlaces The number of decimal places for precision.
     * TODO:
     * Override tag
     */
    //public @Override double getArithmeticMeanDouble( int numberOfDecimalPlaces ) {
    public double getArithmeticMeanDouble( 
            int numberOfDecimalPlaces ) {
        BigDecimal arithmeticMeanBigDecimal = getArithmeticMeanBigDecimal( numberOfDecimalPlaces );
        if ( arithmeticMeanBigDecimal == null ) {
            return Double.NaN;
        } else {
            return arithmeticMeanBigDecimal.doubleValue();
        }
    }
    
    /**
     * For returning the Geometric Mean of all non noDataValues as a double
     * Warning! This is imprecise and it can happen that Math.pow does not
     * return what might be expected! (For example, negative powers in the range
     * (0,1) for negative numbers.)
     * TODO:
     * Develop a pow function such as com.ibm.icu.math.BigDecimal.pow
     * This resource is not used here due to licensing, but it could be...
     */
    public double getGeometricMeanDouble() {
        BigDecimal mean = new BigDecimal(1.0d);
        BigDecimal count = new BigDecimal(0.0d);
        BigDecimal thisCount;
        //double mean = 1.0d;
        //double count = 0.0d;
        //byte thisCount = 0;
        TDoubleLongIterator iterator = this.data.iterator();
        double key;
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            key = iterator.key();
            value.setLong( iterator.value() );
            thisCount = new BigDecimal( ( double ) value.getCount() );
            count = count.add( thisCount );
            mean = mean.multiply( new BigDecimal( key ).multiply( thisCount ) );
            //thisCount = value.getCount();
            //count += thisCount;
            //mean *= key * ( double ) thisCount;
        }
        //return ( mean.pow( new BigDecimal( 1.0d ).divide( count, BigDecimal.ROUND_HALF_EVEN ) ) ).doubleValue();
        return Math.pow( mean.doubleValue(), 1.0d / count.doubleValue() );
        //return Math.pow( mean, 1.0d / count );
    }
    
    /**
     * Returns the Harmonic Mean of all non noDataValues as a double. Zero
     * values are ignored in the calculation.
     */
    public double getHarmonicMeanDouble() {
        BigDecimal mean = new BigDecimal(1.0d);
        BigDecimal count = new BigDecimal(0.0d);
        BigDecimal thisCount;
        BigDecimal one = new BigDecimal(1.0d);
        TDoubleLongIterator iterator = this.data.iterator();
        UnsignedLong value = new UnsignedLong();
        BigDecimal bigDecimal0;
        BigDecimal bigDecimal1;
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = BigDecimal.valueOf( ( long ) value.getCount() );
            if ( iterator.key() != 0.0d ) {
                count = count.add( thisCount );
                bigDecimal0 = new BigDecimal( iterator.key() );
                bigDecimal1 = one.divide( 
                        bigDecimal0, 
                        325, 
                        BigDecimal.ROUND_HALF_EVEN );
                mean = mean.add( thisCount.multiply( bigDecimal1 ) );
            } else {
                System.out.println(
                        "Warning: Value of 0.0d encountered in " + 
                        this.getClass().getName() + ".getHarmonicMeanDouble()!" );
            }
        }
        if ( mean.compareTo( BigDecimal.ZERO ) != 0 ) {
            bigDecimal0 = count.divide( 
                    mean,
                    325,
                    BigDecimal.ROUND_HALF_EVEN );
            return bigDecimal0.doubleValue();
        } else {
            return mean.doubleValue();
        }
    }
    
    /**
     * For returning the mode of all non noDataValues as a TDoubleHashSet
     */
    public @Override TDoubleHashSet getModeTDoubleHashSet() {
        TDoubleHashSet mode = new TDoubleHashSet();
        TDoubleLongIterator iterator = this.data.iterator();
        byte modeCount = 0;
        byte thisCount;
        UnsignedLong value = new UnsignedLong();
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = value.getCount();
            if ( thisCount > modeCount ) {
                mode.clear();
                mode.add( iterator.key() );
                modeCount = thisCount;
            } else {
                if ( thisCount == modeCount ) {
                    mode.add( iterator.key() );
                }
            }
        }
        return mode;
    }
    
    /**
     * For returning the median of all non noDataValues as a double
     */
    public @Override double getMedianDouble() {
        int scale = 325;
        double[] keys = this.data.keys();
        sort1( keys, 0, keys.length );
        int nonNoDataValueCount = getNonNoDataValueCountInt();
        UnsignedLong unsignedLong = new UnsignedLong();
        int index = -1;
        if ( nonNoDataValueCount % 2 == 0 ) {
            // Need arithmetic mean of ( ( nonNoDataValueCount / 2 ) - 1 )th
            // and ( nonNoDataValueCount / 2 )th values
            int requiredIndex = ( nonNoDataValueCount / 2 ) - 1;
            boolean got = false;
            //double median = 0.0d;
            BigDecimal medianBigDecimal = null;
            for ( int keyIndex = 0; keyIndex < keys.length; keyIndex ++ ) {
                unsignedLong.setLong( this.data.get( keys[ keyIndex ] ) );
                index += unsignedLong.getCount();
                if ( ! got && index >= requiredIndex ) {
                    //median = keys[ keyIndex ];
                    medianBigDecimal = new BigDecimal( keys[ keyIndex ] );
                    got = true;
                }
                if ( index >= requiredIndex + 1 ) {
                    //median += keys[ keyIndex ];
                    //median /= 2.0d;
                    //return ( median + keys[ keyIndex ] ) / 2.0d;
                    BigDecimal bigDecimal0 = new BigDecimal( keys[ keyIndex ] );
                    BigDecimal bigDecimal1 = medianBigDecimal.add( bigDecimal0 );
                    BigDecimal bigDecimalTwo = new BigDecimal( 2.0d );
                    bigDecimal0 = ( bigDecimal1 ).divide( 
                            bigDecimalTwo, 
                            scale, 
                            BigDecimal.ROUND_HALF_EVEN );
                    return bigDecimal0.doubleValue();
                }
            }
        } else {
            // Need ( ( nonNoDataValueCount ) / 2 )th value
            int requiredIndex = nonNoDataValueCount / 2;
            for ( int keyIndex = 0; keyIndex < keys.length; keyIndex ++ ) {
                unsignedLong.setLong( this.data.get( keys[ keyIndex ] ) );
                index += unsignedLong.getCount();
                if ( index >= requiredIndex ) {
                    return keys[ keyIndex ];
                }
            }
        }
        return getGrid2DSquareCellDouble().noDataValue;
    }
    
    /**
     * For returning the standard deviation of all non noDataValues as a double
     */
    //public @Override double getStandardDeviationDouble( int numberOfDecimalPlaces ) {
    public double getStandardDeviationDouble( int numberOfDecimalPlaces ) {
        double standardDeviation = 0.0d;
        double mean = getArithmeticMeanDouble( numberOfDecimalPlaces );
        TDoubleLongIterator iterator = this.data.iterator();
        UnsignedLong value = new UnsignedLong();
        double count = 0.0d;
        double thisCount;
        for ( int ite = 0; ite < data.size(); ite ++ ) {
            iterator.advance();
            value.setLong( iterator.value() );
            thisCount = ( double ) value.getCount();
            count += thisCount;
            standardDeviation += 
                    ( ( iterator.key() - mean ) * ( iterator.key() - mean ) )
                    * thisCount;
        }
        if ( ( count - 1.0d ) > 0 ) {
            return Math.sqrt( standardDeviation / ( count - 1.0d ) );
        } else {
            return standardDeviation;
        }
    }
    
    /**
     * For returning the number of different values.
     */
    public BigInteger getDiversity() {
        return new BigInteger( Integer.toString( this.data.size() ) );
    }
    
    /**
     * Returns a Grid2DSquareCellDoubleChunk64CellMapIterator for iterating over
     * the cells in this.
     */
    public IteratorAbstract iterator() {
        return new Grid2DSquareCellDoubleChunk64CellMapIterator( this );
    }
    
}