/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import gnu.trove.TDoubleHashSet;
import gnu.trove.TDoubleObjectHashMap;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashSet;
import java.util.Iterator;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
/**
 * Provides general methods and controls what methods extended classes must 
 * implement acting as an interface.
 */
public abstract class Grid2DSquareCellDoubleChunkAbstract 
        extends Grid2DSquareCellChunkAbstract 
        implements Serializable {
        //implements Serializable, GridDoubleStatisticsInterface {
    
    private static final long serialVersionUID = 1L; 
    
    /**
     * Returns grid2DSquareCellDouble.
     */
    public Grid2DSquareCellDouble getGrid2DSquareCellDouble() {
        return ( Grid2DSquareCellDouble ) this.grid2DSquareCell;
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex;
     * chunk cell row chunkCellColIndex, as a double.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param noDataValue the noDataValue of this.grid2DSquareCellDouble
     */
    public abstract double getCell(
            int chunkCellRowIndex,
            int chunkCellColIndex, 
            double noDataValue );
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex;
     * chunk cell row chunkCellColIndex, as a BigDecimal. If the value is this.noDataValue 
     * then null is returned!
     * @param chunkCellRowIndex The row index of the cell w.r.t. the origin of this chunk.
     * @param chunkCellColIndex The column index of the cell w.r.t. the origin of this chunk.
     * @param noDataValue The noDataValue of this.grid2DSquareCellDouble.
     */
    public BigDecimal getCellBigDecimal(
            int chunkCellRowIndex,
            int chunkCellColIndex,
            double noDataValue ) {
        double value = getCell( 
                chunkCellRowIndex, 
                chunkCellColIndex, 
                noDataValue );
        try {
            return new BigDecimal( value );
        } catch ( ArithmeticException ae0 ) {
            return null;
        }
    }
    
    /**
     * Initialises the value at position given by: chunk cell row chunkCellRowIndex;
     * chunk cell column chunkCellColIndex. Utility method for constructors of
     * extending classes.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToInitialise the value with which the cell is initialised
     */
    protected abstract void initCell(
            int chunkCellRowIndex, 
            int chunkCellColIndex,
            double valueToInitialise );
    
    /**
     * Initialises the value of this with those in the randomAccessFiles. The
     * initialisation is done in row order and ifthe end of the file is reached
     * then the remaining values are assumed to be noDataValues.
     * NB1. This method is intended mainly for use in caching
     */
    protected void initCells( 
            Grid2DSquareCellDoubleChunkRAF grid2DSquareCellDoubleChunkRAF ) {
        try {
            Grid2DSquareCellDouble grid2DSquareCellDouble = 
                    getGrid2DSquareCellDouble();
            int chunkNrows = grid2DSquareCellDouble.chunkNrows;
            int chunkNcols = grid2DSquareCellDouble.chunkNcols;
            RandomAccessFile randomAccessFile = 
                    grid2DSquareCellDoubleChunkRAF.randomAccessFile;
            randomAccessFile.seek(0L);
            int row;
            int col;
            double noDataValue = grid2DSquareCellDouble.noDataValue;
            for ( row = 0; row < chunkNrows; row ++ ) {
                for ( col = 0; col < chunkNcols; col ++ ) {
                    try {
                        initCell( 
                                row, 
                                col, 
                                randomAccessFile.readDouble() );
                    } catch ( IOException ioe1 ) {
                        initCell( 
                                row, 
                                col, 
                                noDataValue );
                    }
                }
            }
        } catch ( IOException ioe0 ) {
            ioe0.printStackTrace();
        }
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex;
     * chunk cell row chunkCellColIndex and sets it to valueToSet
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToSet the value the cell is to be set to
     * @param noDataValue the noDataValue of this.grid2DSquareCellDouble
     */
    public abstract double setCell( 
            int chunkCellRowIndex, 
            int chunkCellColIndex, 
            double valueToSet, 
            double noDataValue );
    
    /**
     * For clearing the data associated with this.
     */
    public abstract void clearData();
    
    /**
     * For initialising the data associated with this.
     */
    public abstract void initData();
    
    /**
     * TODO: docs
     */
    public double[] toArrayIncludingNoDataValues() {
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
        int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
        double[] array;
        if ( ( ( long ) nrows * ( long ) ncols ) > Integer.MAX_VALUE ) {
            //throw new PrecisionExcpetion
            System.out.println(
                    "PrecisionException in " + this.getClass().getName() + 
                    ".toArray()!" );
            System.out.println(
                    "Warning! The returned array size is only " + 
                    Integer.MAX_VALUE + " instead of " + 
                    ( ( long ) nrows * ( long ) ncols ) );
            array = new double[ Integer.MAX_VALUE ];
        }
        double noDataValue = getGrid2DSquareCellDouble().noDataValue;
        array = new double[ nrows * ncols ];
        int row;
        int col;
        int count = 0;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                array[ count ] = getCell( 
                        row, 
                        col, 
                        noDataValue );
                count ++;
            }
        }
        return array;
    }
    
    /**
     * TODO: docs
     */
    public double[] toArrayNotIncludingNoDataValues() {
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
        int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
        double noDataValue = grid2DSquareCellDouble.noDataValue;
        double[] array = new double[ getNonNoDataValueCountInt() ];
        int row;
        int col;
        int count = 0;
        double value;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                value = getCell( 
                        row, 
                        col, 
                        noDataValue );
                if ( value != noDataValue ) {
                    array[ count ] = value;
                    count ++;
                }
            }
        }
        return array;
    }
    
    /**
     * Returns the number of cells with noDataValues as a BigInteger
     */
    public BigInteger getNonNoDataValueCountBigInteger() {
        BigInteger nonNoDataValueCount = BigInteger.ZERO;
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
        int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
        double noDataValue = grid2DSquareCellDouble.noDataValue;
        for ( int row = 0; row < nrows; row ++ ) {
            for ( int col = 0; col < ncols; col ++ ) {
                if ( getCell( row, col, noDataValue ) != noDataValue ) {
                    nonNoDataValueCount = nonNoDataValueCount.add( BigInteger.ONE );
                }
            }
        }
        return nonNoDataValueCount;
    }
    
    /**
     * Returns the number of cells with noDataValues as an int
     */
    public int getNonNoDataValueCountInt() {
        return getNonNoDataValueCountBigInteger().intValue();
    }
    
    /**
     * For returning the sum of all non noDataValues as a BigDecimal
     */
    public BigDecimal getSumBigDecimal() {
        BigDecimal sum = BigDecimal.ZERO;
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
        int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
        double noDataValue = grid2DSquareCellDouble.noDataValue;
        double value;
        int row;
        int col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                value = getCell( row, col, noDataValue );
                if ( value != noDataValue ) {
                    sum = sum.add( new BigDecimal( value ) );
                }
            }
        }
        return sum;
    }
    
    /**
     * For returning the minimum of all non noDataValues as a double
     */
    public double getMinDouble() {
        double min = Double.POSITIVE_INFINITY;
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
       int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
        double noDataValue = grid2DSquareCellDouble.noDataValue;
        double value;
        int row;
        int col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                value = getCell( 
                        row, 
                        col, 
                        noDataValue );
                if ( value != noDataValue ) {
                    min = Math.min( 
                            min, 
                            value );
                }
            }
        }
        return min;
    }
    
    /**
     * For returning the maximum of all non noDataValues as a double
     */
    public double getMaxDouble() {
        double max = Double.NEGATIVE_INFINITY;
        Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
      int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
          double noDataValue = grid2DSquareCellDouble.noDataValue;
        double value;
        int row;
        int col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                value = getCell( 
                        row, 
                        col, 
                        noDataValue );
                if ( value != noDataValue ) {
                    max = Math.max( 
                            max, 
                            value );
                }
            }
        }
        return max;
    }
    
    /**
     * For returning the Arithmetic Mean of all non noDataValues as a BigDecimal. If all
     * cells are noDataValues, then null is returned.
     * @param numberOfDecimalPlaces The number of decimal places to which the result is precise.
     */
    public BigDecimal getArithmeticMeanBigDecimal( 
            int numberOfDecimalPlaces ) {
        BigDecimal mean = BigDecimal.ZERO;
        BigDecimal count = BigDecimal.ZERO;
        BigDecimal one = BigDecimal.ONE;
        Grid2DSquareCellDouble grid2DSquareCellDouble = 
                getGrid2DSquareCellDouble();
     int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
          double noDataValue = grid2DSquareCellDouble.noDataValue;
        double value;
        int row;
        int col;
        for ( row = 0; row < nrows; row ++ ) {
            for ( col = 0; col < ncols; col ++ ) {
                value = getCell( row, col, noDataValue );
                if ( value != noDataValue ) {
                    mean = mean.add( new BigDecimal( value ) );
                    count = count.add( one );
                }
            }
        }
        if ( count.compareTo( BigDecimal.ZERO ) != 0 ) {
            return mean.divide( 
                    count, 
                    numberOfDecimalPlaces, 
                    BigDecimal.ROUND_HALF_EVEN );
        }
        return null;
    }
    
    /**
     * For returning the Arithmetic Mean of all non noDataValues as a double. If all
     * cells are noDataValues, then noDataValue is returned.
     * Using BigDecimal this should be as precise as possible with doubles.
     */
    public double getArithmeticMeanDouble() {
        BigDecimal arithmeticMeanBigDecimal = getArithmeticMeanBigDecimal( 324 );
        try {
            return arithmeticMeanBigDecimal.doubleValue();
        } catch ( NullPointerException npe0 ) {
            return getGrid2DSquareCellDouble().noDataValue;
        }
    }
    
    /**
     * For returning the mode of all non noDataValues as a TDoubleHashSet.
     * TODO: Better to use toArray and go through a sorted version?
     */
    public TDoubleHashSet getModeTDoubleHashSet() {
        TDoubleHashSet mode = new TDoubleHashSet();
        BigInteger nonNoDataValueCountBigInteger = 
                getNonNoDataValueCountBigInteger();
        if ( nonNoDataValueCountBigInteger.compareTo( BigInteger.ZERO ) == 1 ) {
            //TDoubleObjectHashMap modes = new TDoubleObjectHashMap();
            Grid2DSquareCellDouble grid2DSquareCellDouble = 
                    getGrid2DSquareCellDouble();
          int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
           double noDataValue = grid2DSquareCellDouble.noDataValue;
           boolean calculated = false;
            int row = 0;
            int col = 0;
            int p;
            int q;
            Object[] tmode = initMode( nrows, ncols, noDataValue );
            if ( tmode[0] == null ) {
                return mode;
            } else {
                double value;
                long count;
                long modeCount = ( ( Long ) tmode[ 0 ] ).longValue();
                mode.add( ( ( Double ) tmode[ 1 ] ).doubleValue() );
                ChunkCellID chunkCellID = ( ChunkCellID ) tmode[ 2 ];
                // Do remainder of the row
                p = chunkCellID.chunkCellRowIndex;
                for ( q = chunkCellID.chunkCellColIndex + 1; q < ncols; q ++ ) {
                    value = getCell( p, q, noDataValue );
                    if ( value != noDataValue ) {
                        count = count( 
                                p, 
                                q, 
                                nrows,
                                ncols, 
                                value );
                        if ( count > modeCount ) {
                            mode.clear();
                            mode.add( value );
                            modeCount = count;
                        } else {
                            if ( count == modeCount ) {
                                mode.add( value );
                            }
                        }
                    }
                }
                // Do remainder of the grid
                for ( p ++ ; p < nrows; p ++ ) {
                    for ( q = 0; q < ncols; q ++ ) {
                        value = getCell( p, q, noDataValue );
                        if ( value != noDataValue ) {
                            count = count( 
                                    p, 
                                    q, 
                                    nrows, 
                                    ncols, 
                                    value );
                            if ( count > modeCount ) {
                                mode.clear();
                                mode.add( value );
                                modeCount = count;
                            } else {
                                if ( count == modeCount ) {
                                    mode.add( value );
                                }
                            }
                        }
                    }
                }
            }
        }
        return mode;
    }
    
    /**
     * TODO: docs
     */
    private Object[] initMode( 
            int nrows, 
            int ncols, 
            double noDataValue ) {
        Object[] initMode = new Object[3];
        long modeCount;
        int p;
        int q;
        int row;
        int col;
        double value;
        double thisValue;
        for ( p = 0; p < nrows; p ++ ) {
            for ( q = 0; q < ncols; q ++ ) {
                value = getCell( p, q, noDataValue );
                if ( value != noDataValue ) {
                    modeCount = 0L;
                    for ( row = 0; row < nrows; row ++ ) {
                        for ( col = 0; col < ncols; col ++ ) {
                            thisValue = getCell(
                                    row, 
                                    col, 
                                    noDataValue );
                            if ( thisValue == value ) {
                                modeCount ++;
                            }
                        }
                    }
                    initMode[0] = new Long( modeCount );
                    initMode[1] = new Double( value );
                    initMode[2] = new ChunkCellID( 
                            p, 
                            q );
                    return initMode;
                }
            }
        }
        return initMode;
    }
    
    /**
     * TODO: docs
     * @param p the row index of the cell from which counting starts
     * @param q the column index of the cell from which counting starts
     * @param nrows
     * @param ncols
     * @param value the value to be counted
     */
    private long count( 
            int p,
            int q,
            int nrows, 
            int ncols, 
            double value ) {
        long count = 1L;
        double thisValue;
        double noDataValue = 
                getGrid2DSquareCellDouble().noDataValue;
        // Do remainder of the row
        for ( q ++; q < ncols; q ++ ) {
            thisValue = getCell( 
                    p, 
                    q, 
                    noDataValue );
            if ( thisValue == value ) {
                count ++;
            }
        }
        // Do remainder of the grid
        for ( p ++ ; p < nrows; p ++ ) {
            for ( q = 0; q < ncols; q ++ ) {
                thisValue = getCell( 
                        p, 
                        q, 
                        noDataValue );
                if ( thisValue == value ) {
                    count ++;
                }
            }
        }
        return count;
    }
    
    /**
     * For returning the median of all non noDataValues as a double. This method
     * requires that all data in chunk can be stored as a new array.
     */
    public double getMedianDouble() {
        try {
            Grid2DSquareCellDouble grid2DSquareCellDouble = 
                    getGrid2DSquareCellDouble();
            int scale = 20;
            double median = grid2DSquareCellDouble.noDataValue;
            BigInteger nonNoDataValueCountBigInteger = 
                    getNonNoDataValueCountBigInteger();
            if ( nonNoDataValueCountBigInteger.compareTo( BigInteger.ZERO ) == 1 ) {
                double[] array = toArrayNotIncludingNoDataValues();
                sort1( array, 0, array.length );
                BigInteger[] nonNoDataValueCountBigIntegerDivideAndRemainder2 = 
                        nonNoDataValueCountBigInteger.divideAndRemainder( new BigInteger( "2" ) );
                if ( nonNoDataValueCountBigIntegerDivideAndRemainder2[ 1 ].compareTo( BigInteger.ZERO ) == 0 ) {
                    int index = nonNoDataValueCountBigIntegerDivideAndRemainder2[ 0 ].intValue();
                    //median = array[ index ];
                    //median += array[ index - 1 ];
                    //median /= 2.0d;
                    //return median;
                    BigDecimal medianBigDecimal = new BigDecimal( array[ index - 1 ] );
                    return ( medianBigDecimal.add( new BigDecimal( array[ index ] ) ) ).divide( new BigDecimal( 2.0d ), scale, BigDecimal.ROUND_HALF_DOWN ).doubleValue();
                    //return ( medianBigDecimal.add( new BigDecimal( array[ index ] ) ) ).divide( new BigDecimal( 2.0d ), scale, BigDecimal.ROUND_HALF_EVEN ).doubleValue();
                } else {
                    int index = nonNoDataValueCountBigIntegerDivideAndRemainder2[ 0 ].intValue();
                    return array[ index ];
                }
            } else {
                return median;
            }
        } catch ( OutOfMemoryError oome0 ) {
            Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
            grid2DSquareCellDouble.clearMemoryReserve();
            grid2DSquareCellDouble.writeToFileGrid2DSquareCellChunksExcept( 
                    this.chunkID, 
                    false );
            grid2DSquareCellDouble.clearFromCacheGrid2DSquareCellChunksExcept( 
                    this.chunkID, false );
            grid2DSquareCellDouble.initMemoryReserve();
            double median = grid2DSquareCellDouble.noDataValue;
            // Try to cache all other chunks and calculate median again.
            // (N.B. If this is the only chunk not cached this will fail again.)
            System.out.println("Warning: Ran out of memory calculating median. Trying to cache out some other chunk and try again. " );
            try {
                int scale = 325;
                BigInteger nonNoDataValueCountBigInteger = getNonNoDataValueCountBigInteger();
                if ( nonNoDataValueCountBigInteger.compareTo( BigInteger.ZERO ) == 1 ) {
                    double[] array = toArrayNotIncludingNoDataValues();
                    sort1( array, 0, array.length );
                    BigInteger[] nonNoDataValueCountBigIntegerDivideAndRemainder2 = nonNoDataValueCountBigInteger.divideAndRemainder( new BigInteger( "2" ) );
                    if ( nonNoDataValueCountBigIntegerDivideAndRemainder2[ 1 ].compareTo( BigInteger.ZERO ) == 0 ) {
                        int index = nonNoDataValueCountBigIntegerDivideAndRemainder2[ 0 ].intValue();
                        //median = array[ index ];
                        //median += array[ index - 1 ];
                        //median /= 2.0d;
                        //return median;
                        BigDecimal medianBigDecimal = new BigDecimal( array[ index - 1 ] );
                        return ( medianBigDecimal.add( new BigDecimal( array[ index ] ) ) ).divide( new BigDecimal( 2.0d ), scale, BigDecimal.ROUND_HALF_DOWN ).doubleValue();
                        //return ( medianBigDecimal.add( new BigDecimal( array[ index ] ) ) ).divide( new BigDecimal( 2.0d ), scale, BigDecimal.ROUND_HALF_EVEN ).doubleValue();
                    } else {
                        int index = nonNoDataValueCountBigIntegerDivideAndRemainder2[ 0 ].intValue();
                        return array[ index ];
                    }
                } else {
                    return median;
                }
            } catch ( OutOfMemoryError oome1 ) {
                try {
                    grid2DSquareCellDouble.clearMemoryReserve();
                    if ( ! ( this instanceof Grid2DSquareCellDoubleChunkRAF ) ) {
                        Grid2DSquareCellDoubleChunkRAFFactory grid2DSquareCellDoubleChunkRAFFactory = new Grid2DSquareCellDoubleChunkRAFFactory();
                        Grid2DSquareCellDoubleChunkRAF grid2DSquareCellDoubleChunkRAF = ( Grid2DSquareCellDoubleChunkRAF ) grid2DSquareCellDoubleChunkRAFFactory.createGrid2DSquareCellDoubleChunk( this, this.chunkID );
                        clearData();
                        grid2DSquareCellDouble.initMemoryReserve();
                        int scale = 325;
                        BigInteger nonNoDataValueCountBigInteger = grid2DSquareCellDoubleChunkRAF.getNonNoDataValueCountBigInteger();
                        if ( nonNoDataValueCountBigInteger.compareTo( BigInteger.ZERO ) == 1 ) {
                            double[] array = grid2DSquareCellDoubleChunkRAF.toArrayNotIncludingNoDataValues();
                            sort1( array, 0, array.length );
                            BigInteger[] nonNoDataValueCountBigIntegerDivideAndRemainder2 = nonNoDataValueCountBigInteger.divideAndRemainder( new BigInteger( "2" ) );
                            if ( nonNoDataValueCountBigIntegerDivideAndRemainder2[ 1 ].compareTo( BigInteger.ZERO ) == 0 ) {
                                int index = nonNoDataValueCountBigIntegerDivideAndRemainder2[ 0 ].intValue();
                                //median = array[ index ];
                                //median += array[ index - 1 ];
                                //median /= 2.0d;
                                //return median;
                                BigDecimal medianBigDecimal = new BigDecimal( array[ index - 1 ] );
                                median = ( medianBigDecimal.add( new BigDecimal( array[ index ] ) ) ).divide( new BigDecimal( 2.0d ), scale, BigDecimal.ROUND_HALF_DOWN ).doubleValue();
                                //return ( medianBigDecimal.add( new BigDecimal( array[ index ] ) ) ).divide( new BigDecimal( 2.0d ), scale, BigDecimal.ROUND_HALF_EVEN ).doubleValue();
                            } else {
                                int index = nonNoDataValueCountBigIntegerDivideAndRemainder2[ 0 ].intValue();
                                median = array[ index ];
                            }
                        }
                        // Reload this from grid2DSquareCellDoubleChunkRAF
                        System.gc();
                        initData();
                        initCells( grid2DSquareCellDoubleChunkRAF );
                        // Set the chunk in grid to be grid2DSquareCellDoubleChunkRAF
                        //grid2DSquareCellDoubleChunkRAF.initGrid2DSquareCellDouble( grid2DSquareCellDouble );
                        //grid2DSquareCellDouble.setChunk( grid2DSquareCellDoubleChunkRAF, this.chunkID );
                    }
                    grid2DSquareCellDouble.initMemoryReserve();
                    return median;
                } catch ( OutOfMemoryError oome2 ) {
                    System.out.println("Warning: Unable to calculate median. Returning the mean instead!");
                    return getArithmeticMeanDouble();
                    // Code incomplete. Basically what is needed is a sort that
                    // works with arrays on files rather than with arrays in memory.
                    // The solution is likely to be the way getMedian() should be
                    // written for the entire grid.
                    //double median;
                    //// If numberOfNonNoDataValues in int range.
                    //if ( numberOfNonNoDataValues.compareTo( new BigInteger( Integer.toString( Integer.MAX_VALUE ) ) ) == -1 ) {
                    //    Grid2DSquareCellDouble grid2DSquareCellDouble = getGrid2DSquareCellDouble();
                    //    int nrows = grid2DSquareCellDouble.getChunkNrows();
                    //    int ncols = grid2DSquareCellDouble.getChunkNrows();
                    //    double noDataValue = grid2DSquareCellDouble.getNoDataValue();
                    //    double mean = getArithmeticMean();
                    //    int div = 100;
                    //    int row;
                    //    int col;
                    //    double value;
                    //    int lessThanMeanCount = 0;
                    //    int meanCount = 0;
                    //    int greaterThanMeanCount = 0;
                    //    for ( row = 0; row < nrows; row ++ ) {
                    //        for ( col = 0; col < ncols; col ++ ) {
                    //            value = getCell( row, col );
                    //            if ( value != noDataValue ) {
                    //                if ( value < mean ) {
                    //                    lessThanMeanCount ++;
                    //                } else {
                    //                    if ( value == mean ) {
                    //                        greaterThanMeanCount ++;
                    //                    } else {
                    //                        meanCount ++;
                    //                    }
                    //                }
                    //            }
                    //        }
                    //    }
                    //    return median;
                    //}
                    //// If numberOfNonNoDataValues in long range.
                    //if ( numberOfNonNoDataValues.compareTo( new BigInteger( Long.toString( Long.MAX_VALUE ) ) ) == -1 ) {
                    //    return median;
                    //}
                    //BigInteger[] numberOfNonNoDataValuesDivideAndRemainder2 = numberOfNonNoDataValues.divideAndRemainder( new BigInteger( "2" ) );
                    //if ( numberOfNonNoDataValuesDivideAndRemainder2[ 1 ] == BigInteger.ZERO ) {
                    //}
                }
            }
        }
    }
    
    /**
     * Sorts the specified sub-array of doubles into ascending order.
     * Source copied from java.util.Arrays and method changed so not static for
     * performance reasons.
     */
    protected void sort1(double x[], int off, int len) {
        // Insertion sort on smallest arrays
        if (len < 7) {
            for (int i=off; i<len+off; i++)
                for (int j=i; j>off && x[j-1]>x[j]; j--)
                    swap(x, j, j-1);
            return;
        }
        
        // Choose a partition element, v
        int m = off + (len >> 1);       // Small arrays, middle element
        if (len > 7) {
            int l = off;
            int n = off + len - 1;
            if (len > 40) {        // Big arrays, pseudomedian of 9
                int s = len/8;
                l = med3(x, l,     l+s, l+2*s);
                m = med3(x, m-s,   m,   m+s);
                n = med3(x, n-2*s, n-s, n);
            }
            m = med3(x, l, m, n); // Mid-size, med of 3
        }
        double v = x[m];
        
        // Establish Invariant: v* (<v)* (>v)* v*
        int a = off, b = a, c = off + len - 1, d = c;
        while(true) {
            while (b <= c && x[b] <= v) {
                if (x[b] == v)
                    swap(x, a++, b);
                b++;
            }
            while (c >= b && x[c] >= v) {
                if (x[c] == v)
                    swap(x, c, d--);
                c--;
            }
            if (b > c)
                break;
            swap(x, b++, c--);
        }
        
        // Swap partition elements back to middle
        int s, n = off + len;
        s = Math.min(a-off, b-a  );  vecswap(x, off, b-s, s);
        s = Math.min(d-c,   n-d-1);  vecswap(x, b,   n-s, s);
        
        // Recursively sort non-partition-elements
        if ((s = b-a) > 1)
            sort1(x, off, s);
        if ((s = d-c) > 1)
            sort1(x, n-s, s);
    }
    
    /**
     * Swaps x[a] with x[b].
     * Source copied from java.util.Arrays and method changed so not static for
     * performance reasons.
     */
    private void swap(double x[], int a, int b) {
        double t = x[a];
        x[a] = x[b];
        x[b] = t;
    }
    
    /**
     * Swaps x[a .. (a+n-1)] with x[b .. (b+n-1)].
     * Source copied from java.util.Arrays and method changed so not static for
     * performance reasons.
     */
    private void vecswap(double x[], int a, int b, int n) {
        for (int i=0; i<n; i++, a++, b++)
            swap(x, a, b);
    }
    
    /**
     * Returns the index of the median of the three indexed doubles.
     * Source copied from java.util.Arrays and method changed so not static for
     * performance reasons.
     */
    private int med3(double x[], int a, int b, int c) {
        return (x[a] < x[b] ?
        (x[b] < x[c] ? b : x[a] < x[c] ? c : a) :
            (x[b] > x[c] ? b : x[a] > x[c] ? c : a));
    }
    
    /**
     * For returning the standard deviation of all non noDataValues as a double
     */
    public double getStandardDeviationDouble() {
        double standardDeviation = 0.0d;
        double mean = getArithmeticMeanDouble();
        Grid2DSquareCellDouble grid2DSquareCellDouble = 
                getGrid2DSquareCellDouble();
      int nrows = grid2DSquareCellDouble.chunkNrows;
        int ncols = grid2DSquareCellDouble.chunkNcols;
         double noDataValue = grid2DSquareCellDouble.noDataValue;
        double value;
        double count = 0.0d;
        for ( int row = 0; row < nrows; row ++ ) {
            for ( int col = 0; col < ncols; col ++ ) {
                value = getCell( row, col, noDataValue );
                if ( value != noDataValue ) {
                    standardDeviation += 
                            ( value - mean ) * 
                            ( value - mean );
                    count += 1.0d;
                }
            }
        }
        if ( ( count - 1.0d ) > 0.0d ) {
            return Math.sqrt( standardDeviation / ( count - 1.0d ) );
        } else {
            return standardDeviation;
        }
    }
    
    //    /**
    //     * For returning the number of different values.
    //     */
    //    public BigInteger getDiversity() {
    //        return new BigInteger( Integer.toString( data.size() ) );
    //    }
    
    
    

}