/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import gnu.trove.TDoubleIntHashMap;
import gnu.trove.TDoubleIntIterator;
import gnu.trove.TDoubleIntProcedure;
import gnu.trove.TDoubleLongHashMap;
import gnu.trove.TIntProcedure;
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Arrays;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;
/**
 * Grid2DSquareCellDoubleChunkAbstract extension that stores cell values in a 
 * double[][].
 */
public class Grid2DSquareCellDoubleChunkArray 
        extends Grid2DSquareCellDoubleChunkAbstract 
        implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    /**
     * For storing values arranged in rows and columns.
     */
    private double[][] data;
    
    /**
     * Default constructor
     */
    public Grid2DSquareCellDoubleChunkArray() {
        this.chunkID = new ChunkID();
        this.data = new double[ 1 ][ 1 ];
        this.grid2DSquareCell = new Grid2DSquareCellDouble();
        this.isSwapUpToDate = false;
    }
 
    /**
     * Creates a new Grid2DSquareCellDouble grid containing all no data values.
     * @param grid2DSquareCellDouble
     * @param chunkID
     */
    protected Grid2DSquareCellDoubleChunkArray( 
            Grid2DSquareCellDouble grid2DSquareCellDouble,
            ChunkID chunkID ) {
        this.chunkID = chunkID;
        initGrid2DSquareCell( grid2DSquareCellDouble );
        double noDataValue = grid2DSquareCellDouble.noDataValue;
        int chunkNrows = grid2DSquareCellDouble.getChunkNrows( 
                chunkID, 
                grid2DSquareCellDouble.handleOutOfMemoryErrorFalse );
        this.data = new double[ chunkNrows ][ grid2DSquareCellDouble.getChunkNcols(
                chunkID, 
                grid2DSquareCellDouble.handleOutOfMemoryErrorFalse ) ];
        int row;
        for ( row = 0; row < chunkNrows; row ++ ) {
            Arrays.fill( data[ row ], noDataValue );
        }
        this.isSwapUpToDate = false;
    }
    
    /**
     * TODO:
     * 1. docs
     * 2. A fast toArray() method in Grid2DSquareCellDoubleChunkMap could be coded 
     * then a constructor based on an double[] or double[][] might be faster?
     * @param grid2DSquareCellDoubleChunk
     * @param chunkID
     */
    protected Grid2DSquareCellDoubleChunkArray( 
            Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk, 
            ChunkID chunkID ) {
        this.chunkID = chunkID;
        Grid2DSquareCellDouble grid2DSquareCellDouble = 
                grid2DSquareCellDoubleChunk.getGrid2DSquareCellDouble();
        initGrid2DSquareCell( grid2DSquareCellDouble );
        int chunkNrows = grid2DSquareCellDouble.getChunkNrows( 
                chunkID, 
                grid2DSquareCellDouble.handleOutOfMemoryErrorFalse );
        int chunkNcols = grid2DSquareCellDouble.getChunkNcols( 
                chunkID, 
                grid2DSquareCellDouble.handleOutOfMemoryErrorFalse );
        double noDataValue = grid2DSquareCellDouble.noDataValue;
        initData();
        int row;
        int col;
        for ( row = 0; row < chunkNrows; row ++ ) {
            for ( col = 0; col < chunkNcols; col ++ ) {
                this.data[ row ][ col ] = grid2DSquareCellDoubleChunk.getCell( 
                        row, 
                        col, 
                        noDataValue );
                //initCell( row, col, grid2DSquareCellDoubleChunk.getCell( row, col ) );
            }
        }
        this.isSwapUpToDate = false;
    }
    
    /**
     * Returns a description of this.
     */
    public String toString() {
        return "Grid2DSquareCellDoubleChunkArray( ChunkID( " + this.chunkID.toString() + " ) )";
    }
    
    /**
     * Initialises the data associated with this.
     */
    public void initData() {
        Grid2DSquareCellDouble grid2DSquareCellDouble = this.getGrid2DSquareCellDouble();
        int chunkNrows = grid2DSquareCellDouble.chunkNrows;
        int chunkNcols = grid2DSquareCellDouble.chunkNcols;
        this.data = new double[ chunkNrows ][ chunkNcols ];
    }
    
    /**
     * TODO: docs
     */
    public double[][] getData(){
        return this.data;
    }
    
    /**
     * Clears the data associated with this.
     */
    public void clearData() {
        this.data = null;
        System.gc();
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell row chunkCellColIndex.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param noDataValue the noDataValue of this.grid2DSquareCellDouble
     */
    public double getCell(
            int chunkCellRowIndex,
            int chunkCellColIndex,
            double noDataValue ) {
        try {
            return this.data[ chunkCellRowIndex ][ chunkCellColIndex ];
        } catch ( Exception e0 ) {
            return noDataValue;
        }
    }
    
    /**
     * Initialises the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell column chunkCellColIndex. Utility method for constructor.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToInitialise the value with which the cell is initialised
     */
    protected void initCell(
            int chunkCellRowIndex,
            int chunkCellColIndex, 
            double valueToInitialise) {
        this.data[ chunkCellRowIndex ][ chunkCellColIndex ] = valueToInitialise;
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell row chunkCellColIndex and sets it to valueToSet
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToSet the value the cell is to be set to
     * @param noDataValue the noDataValue of this.grid2DSquareCellDouble
     */
    public double setCell( 
            int chunkCellRowIndex, 
            int chunkCellColIndex, 
            double valueToSet, 
            double noDataValue ) {
        try {
            double oldValue = this.data[ chunkCellRowIndex ][ chunkCellColIndex ];
            this.data[ chunkCellRowIndex ][ chunkCellColIndex ] = valueToSet;
            if ( getIsSwapUpToDate() ) { 
                // Optimisation? Want a setCellFast method closer to initCell? 
                // What about an unmodifiable readOnly type chunk?
                if ( valueToSet != oldValue ) {
                    setIsSwapUpToDate( false );
                }
            }
            return oldValue;
        } catch ( Exception e0 ) { // Should not happen! 
            return noDataValue;
        }
    }
    
    /**
     * Returns a Grid2DSquareCellDoubleChunkArrayIterator for iterating over 
     * the cells in this.
     */
    public IteratorAbstract iterator() {
        return new Grid2DSquareCellDoubleChunkArrayIterator( this );
    }
    
}