/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
import java.io.File;
import java.io.ObjectInputStream;
import java.math.BigDecimal;
import java.util.HashSet;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;

/**
 * A factory for constructing Grid2DSquareCellInt instances.
 */
public class Grid2DSquareCellIntFactory 
        extends Grid2DSquareCellAbstractFactory {
    
    /**
     * The Grid2DSquareCellChunkAbstractFactory for creating chunks.
     */
    protected Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory;
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory
     * Defaults:
     * directory to a new one in System.getProperties( "java.io.tmpdir" );
     */
    public Grid2DSquareCellIntFactory() {
        this( FileCreator.createTempFile() );
        //this ( FileCreator.createNewFile() );
    }
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory
     * @param directory A "workspace directory" for storing temporary files and
     *   swapping Grid2DSquareCellDouble data to.
     * Defaults:
     * chunkNrows to 64
     * chunkNcols to 64
     * Grid2DSquareCellDoubleChunkFactory to grid2DSquareCellDoubleChunkArray
     */
    public Grid2DSquareCellIntFactory(
            File directory ) {
        this(   directory, 
                64, 
                64, 
                new Grid2DSquareCellIntChunkArrayFactory() );
    }
    
    /**
     * Creates a new Grid2DSquareCellDoubleFactory
     * @param directory A "workspace directory" for storing temporary files and
     *   caching Grid2DSquareCellIntAbstract data.
     * @param chunkNrows The number of rows chunks have by default.
     * @param chunkNcols The number of columns chunks have by default.
     * @param grid2DSquareCellIntChunkFactory The Grid2DSquareCellIntChunkAbstractFactory
     *   for creating Grid2DSquareCellDoubleChunks
     */
    public Grid2DSquareCellIntFactory(
            File directory,
            int chunkNrows,
            int chunkNcols,
            Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory ) {
        this.directory = directory;
        if ( grid2DSquareCellIntChunkFactory.getClass() == 
                Grid2DSquareCellIntChunk64CellMapFactory.class ) {
            if ( chunkNrows * chunkNcols > 64 ) {
                chunkNrows = 8;
                chunkNcols = 8;
            }
        }
        this.chunkNrows = chunkNrows;
        this.chunkNcols = chunkNcols;
        this.dimensions = new BigDecimal[ 5 ];
        this.dimensions[ 0 ] = new BigDecimal( 1L );
        this.dimensions[ 1 ] = new BigDecimal( 0L );
        this.dimensions[ 2 ] = new BigDecimal( 0L );
        this.dimensions[ 3 ] = new BigDecimal( chunkNcols );
        this.dimensions[ 4 ] = new BigDecimal( chunkNrows );
        this.grid2DSquareCellIntChunkFactory = grid2DSquareCellIntChunkFactory;
        this.grid2DSquareCells = new HashSet();
        this.gridStatistics = new GridStatistics1();
        this.handleOutOfMemoryError = true;
    }
    
    /**
     * Returns a reference to this.grid2DSquareCellIntChunkFactory.
     */
    public Grid2DSquareCellIntChunkAbstractFactory getGrid2DSquareCellIntChunkFactory() {
        return this.grid2DSquareCellIntChunkFactory;
    }
    
    /**
     * Sets this.grid2DSquareCellIntChunkFactory to grid2DSquareCellIntChunkFactory.
     */
    public void setGrid2DSquareCellIntChunkFactory(
            Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory ) {
        this.grid2DSquareCellIntChunkFactory = grid2DSquareCellIntChunkFactory;
    }
    
    /////////////////////////
    // Create from scratch //
    /////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellInt grid with all values as
     * noDataValues.
     * @param directory The directory to be used for storing cached
     * Grid2DSquareCellInt information.
     * @param nrows the Grid2DSquareCellInt nrows.
     * @param ncols the Grid2DSquareCellInt ncols.
     * @param dimensions
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellInt create(
            File directory,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return create( 
                this.gridStatistics, 
                directory, 
                this.grid2DSquareCellIntChunkFactory, 
                nrows, 
                ncols, 
                dimensions, 
                grid2DSquareCells, 
                handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt grid with all values as noDataValues.
     * @param gridStatistics The GridStatisticsAbstract to accompany the
     *   returned grid.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellInt information.
     * @param grid2DSquareCellIntChunkFactory The 
     *   Grid2DSquareCellIntChunkAbstractFactory for creating chunks.
     * @param nrows The Grid2DSquareCellInt nrows.
     * @param ncols The Grid2DSquareCellInt ncols.
     * @param dimensions
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellInt create(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory,
            long nrows,
            long ncols,
            BigDecimal[] dimensions,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return new Grid2DSquareCellInt( 
                gridStatistics, 
                directory, 
                grid2DSquareCellIntChunkFactory, 
                this.chunkNrows, 
                this.chunkNcols, 
                nrows, 
                ncols, 
                dimensions, 
                grid2DSquareCells, 
                handleOutOfMemoryError );
    }
    
    //////////////////////////////////////////////////////
    // Create from an existing Grid2DSquareCellAbstract //
    //////////////////////////////////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from
     * grid2DSquareCell.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellInt information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values
     *   are obtained.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats
     *   values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats
     *   values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats
     *   values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats
     *   values are used.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellInt create(
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return create( 
                this.gridStatistics, 
                directory, 
                grid2DSquareCell, 
                this.grid2DSquareCellIntChunkFactory, 
                startRowIndex, 
                startColIndex, 
                endRowIndex, 
                endColIndex, 
                grid2DSquareCells, 
                handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from
     * grid2DSquareCell.
     * @param gridStatistics The GridStatisticsAbstract for the returned
     * Grid2DSquareCellInt.
     * @param directory The directory to be used for storing cached
     * Grid2DSquareCellInt information.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract from which values
     * are obtained.
     * @param grid2DSquareCellIntChunkAbstractFactory The
     * Grid2DSquareCellIntChunkAbstractFactory used to construct the chunks.
     * @param startRowIndex The topmost row index of grid2DSquareCell thats
     * values are used.
     * @param startColIndex The leftmost column index of grid2DSquareCell thats
     * values are used.
     * @param endRowIndex The bottom row index of the grid2DSquareCell thats
     * values are used.
     * @param endColIndex The rightmost column index of grid2DSquareCell thats
     * values are used.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellInt create(
            GridStatisticsAbstract gridStatistics,
            File directory,
            Grid2DSquareCellAbstract grid2DSquareCell,
            Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkAbstractFactory,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return new Grid2DSquareCellInt( 
                gridStatistics, directory, 
                grid2DSquareCell, 
                grid2DSquareCellIntChunkAbstractFactory, 
                this.chunkNrows, 
                this.chunkNcols, 
                startRowIndex, 
                startColIndex, 
                endRowIndex, 
                endColIndex, 
                grid2DSquareCells, 
                handleOutOfMemoryError );
    }
    
    
    ////////////////////////
    // Create from a File //
    ////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from
     * gridFile.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellInt information.
     * @param gridFile either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellInt to be returned.
     * @param startRowIndex The topmost row index of the grid represented in
     *   gridFile thats values are used.
     * @param startColIndex The leftmost column index of the grid represented in
     *   gridFile thats values are used.
     * @param endRowIndex The bottom row index of the grid represented in
     *   gridFile thats values are used.
     * @param endColIndex The rightmost column index of the grid represented in
     *   gridFile thats values are used.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellInt create(
            File directory,
            File gridFile,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return create( 
                new GridStatistics1(), 
                directory, gridFile, 
                this.grid2DSquareCellIntChunkFactory, 
                startRowIndex, 
                startColIndex, 
                endRowIndex, 
                endColIndex, 
                grid2DSquareCells, 
                handleOutOfMemoryError );
    }
    
    /**
     * Returns a new Grid2DSquareCellAbstract with values obtained from
     * gridFile.
     * @param directory The directory to be used for storing cached
     *   Grid2DSquareCellInt information.
     * @param gridFile either a directory, or a formatted File with a specific
     *   extension containing the data and information about the
     *   Grid2DSquareCellAbstract to be returned.
     * @param grid2DSquareCellIntChunkFactory The
     *   Grid2DSquareCellIntChunkAbstractFactory used to construct the chunks.
     * @param startRowIndex The topmost row index of the grid represented in
     *   gridFile thats values are used.
     * @param startColIndex The leftmost column index of the grid represented in
     *   gridFile thats values are used.
     * @param endRowIndex The bottom row index of the grid represented in
     *   gridFile thats values are used.
     * @param endColIndex The rightmost column index of the grid represented in
     *   gridFile thats values are used.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public Grid2DSquareCellInt create(
            GridStatisticsAbstract gridStatistics,
            File directory,
            File gridFile,
            Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory,
            long startRowIndex,
            long startColIndex,
            long endRowIndex,
            long endColIndex,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return new Grid2DSquareCellInt( 
                gridStatistics, 
                directory, 
                gridFile, 
                grid2DSquareCellIntChunkFactory, 
                this.chunkNrows, 
                this.chunkNcols, 
                startRowIndex, 
                startColIndex,
                endRowIndex, 
                endColIndex, 
                grid2DSquareCells, 
                handleOutOfMemoryError );
    }
    
    /////////////////////////
    // Create from a cache //
    /////////////////////////
    
    /**
     * Returns a new Grid2DSquareCellInt with values obtained from gridFile.
     * @param directory The directory for swapping to file.
     * @param gridFile A file containing the data to be used in construction.
     * @param ois The ObjectInputStream to construct from.
     * @param grid2DSquareCells A HashSet of swappable Grid2DSquareCellAbstract
     *   instances.
     * @param handleOutOfMemoryError If true then OutOfMemoryErrors are caught
     *   in this method then swap operations are initiated prior to retrying.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public @Override Grid2DSquareCellInt create(
            File directory,
            File gridFile,
            ObjectInputStream ois,
            HashSet grid2DSquareCells,
            boolean handleOutOfMemoryError ) {
        return new Grid2DSquareCellInt( 
                directory, 
                gridFile, 
                ois, 
                grid2DSquareCells, 
                handleOutOfMemoryError );
    }
    
}
